@@ -14,7 +14,7 @@
 
 #' @title Functions for Querying, Validating and Extracting from Formulas
 #'
-#' A suite of utilities for handling model formulas of the style used in Statnet packages.
+#' @description A suite of utilities for handling model formulas of the style used in Statnet packages.
 #'
 #' @name formula.utilities
 NULL
@@ -26,14 +26,17 @@ NULL
 #' \code{keep.onesided==FALSE} (the default).
 #' 
 #' @param object formula object to be updated or evaluated
-#' @param newterms list of terms (names) to append to the formula, or
-#'   a formula whose RHS terms will be used; either may have a "sign"
-#'   attribute vector of the same length as the list, giving the sign
-#'   of each term (`+1` or `-1`).
+#' @param newterms a [`term_list`] object, or any list of terms (names
+#'   or calls) to append to the formula, or a formula whose RHS terms
+#'   will be used; its `"sign"` attribute vector can give the sign of
+#'   each term (`+1` or `-1`), and its `"env"` attribute
+#'   vector will be used to set its environment, with the first
+#'   available being used and subsequent ones producing a warning.
 #' @param keep.onesided if the initial formula is one-sided, keep it
 #'   whether to keep it one-sided or whether to make the initial
 #'   formula the new LHS
-#' @param env an environment for the new formula, if `object` is `NULL`
+#' @param env an environment for the new formula, if `object` is
+#'   `NULL`
 #' @param \dots Additional arguments. Currently unused.
 #' @return \code{append_rhs.formula} each return an updated formula
 #'   object; if `object` is `NULL` (the default), a one-sided formula
@@ -50,6 +53,13 @@ NULL
 #' (f6 <- append_rhs.formula(NULL,list(as.name("z"))))
 #' (f7 <- append_rhs.formula(NULL,structure(list(as.name("z")),sign=-1)))
 #'
+#' fe <- ~z2+z3
+#' environment(fe) <- new.env()
+#' (f8 <- append_rhs.formula(NULL, fe)) # OK
+#' (f9 <- append_rhs.formula(y~x, fe)) # Warning
+#' (f10 <- append_rhs.formula(y~x, fe, env=NULL)) # No warning, environment from fe.
+#' (f11 <- append_rhs.formula(fe, ~z1)) # Warning, environment from fe
+#'
 #' \dontshow{
 #' stopifnot(f1 == (y~x+z1+z2))
 #' stopifnot(f2 == (y~z))
@@ -58,10 +68,15 @@ NULL
 #' stopifnot(f5 == (y~x+z1-z2))
 #' stopifnot(f6 == (~z))
 #' stopifnot(f7 == (~-z))
+#' stopifnot(f8 == (~z2+z3), identical(environment(f8), environment(fe)))
+#' stopifnot(f9 == (y~x+z2+z3), identical(environment(f9), globalenv()))
+#' stopifnot(f10 == (y~x+z2+z3), identical(environment(f10), environment(fe)))
+#' stopifnot(f11 == (z2+z3~z1), identical(environment(f11), environment(fe)))
 #' }
 #'
 #' @export
-append_rhs.formula <- function(object = NULL, newterms, keep.onesided = FALSE, env = environment(object)){
+append_rhs.formula <- function(object = NULL, newterms, keep.onesided = FALSE, env = if(is.null(object)) NULL else environment(object)){
+  force(env)
   if(is.null(object)) keep.onesided <- TRUE
 
   if(inherits(newterms,"formula")) newterms <- list_rhs.formula(newterms)
@@ -71,11 +86,17 @@ append_rhs.formula <- function(object = NULL, newterms, keep.onesided = FALSE, e
     termsign <- if(NVL(attr(newterms, "sign")[i], +1)>0) "+" else "-"
     if(length(object)==0){
       if(termsign == "-") newterm <- call(termsign, newterm)
-      object <- as.formula(call("~", newterm), env = env)
+      object <- as.formula(call("~", newterm))
     }else if(length(object)==3) object[[3L]]<-call(termsign,object[[3L]],newterm)
     else if(keep.onesided) object[[2L]]<-call(termsign,object[[2L]],newterm)
     else object[[3L]]<- if(termsign=="+") newterm else call(termsign,newterm)
+
+    NVL(env) <- termenv <- attr(newterms, "env")[[i]]
+    if(!is.null(termenv) && !identical(env, termenv))
+      warning(sQuote(paste0("newterms[[",i,"]]")), " has an environment that differs from the specified environment or another term's environment.")
   }
+
+  environment(object) <- env
   object
 }
 
@@ -238,27 +259,116 @@ nonsimp.update.formula<-function (object, new, ..., from.new=FALSE){
   nonsimp_update.formula(object, new, ..., from.new=from.new)
 }
 
+#' A helper class for list of terms in an formula
+#'
+#' Typically generated by [list_rhs.formula()], it contains, in
+#' addition to a list of [call()] or similar objects, attributes
+#' `"sign"` and `"env"`, containing, respectively a vector of
+#' signs that the terms had in the original formula and a list of
+#' environments of the formula from which the term has been
+#' extracted. Indexing and concatenation methods preserve these.
+#'
+#' @param x a list of terms or a term; a `term_list`
+#' @param sign a vector specifying the signs associated with each term (`-1` and `+1`)
+#' @param env a list specifying the environments, or NULL
+#' @param i list index
+#' @param ... additional arguments to methods
+#'
+#' @seealso [list_rhs.formula()], [list_summands.call()]
+#'
+#' @examples
+#'
+#' e1 <- new.env()
+#' f1 <- a~b+c
+#' environment(f1) <- e1
+#' f2 <- ~-NULL+1
+#'
+#' (l1 <- list_rhs.formula(f1))
+#' (l2 <- list_rhs.formula(f2))
+#'
+#' (l <- c(l1,l2))
+#' \dontshow{
+#' stopifnot(identical(c(unclass(l)), alist(b, c, NULL, 1)))
+#' stopifnot(identical(attr(l, "sign"), c(1,1,-1,1)))
+#' stopifnot(identical(attr(l, "env"), rep(list(e1, globalenv()), each=2)))
+#' }
+#'
+#' (l <- c(l2[1], l1[2], l1[1], l1[1], l2[2]))
+#' \dontshow{
+#' stopifnot(identical(c(unclass(l)), alist(NULL, c, b, b, 1)))
+#' stopifnot(identical(attr(l, "sign"), c(-1,1,1,1,1)))
+#' stopifnot(identical(attr(l, "env"), list(globalenv(), e1, e1, e1, globalenv())))
+#' }
+#'
+#' @export
+term_list <- function(x, sign = +1, env = NULL){
+  if(!is.list(x)) x <- list(x)
+  if(!is.list(env)) env <- list(env)
+  structure(x,
+            sign = rep(sign, length.out=length(x)),
+            env = rep(env, length.out=length(x)),
+            class = "term_list")
+}
+
+#' @rdname term_list
+#' @export
+as.term_list <- function(x, ...) UseMethod("as.term_list")
+
+#' @rdname term_list
+#' @export
+as.term_list.term_list <- function(x, ...) x
+
+#' @rdname term_list
+#' @export
+as.term_list.default <- function(x, sign = +1, env = NULL, ...) term_list(x, sign=sign, env=env)
+
+#' @rdname term_list
+#' @export
+c.term_list <- function(x, ...){
+  xl <- c(list(as.term_list(x)), lapply(list(...), as.term_list))
+  structure(
+    c(unclass(x), ...),
+    sign = unlist(lapply(xl, attr, "sign"), use.names=FALSE),
+    env = unlist(lapply(xl, attr, "env"), recursive=FALSE, use.names=FALSE),
+    class = "term_list"
+  )
+}
+
+#' @rdname term_list
+#' @export
+`[.term_list` <- function(x, i, ...){
+  term_list(NextMethod(), sign = attr(x, "sign")[i], env = attr(x, "env")[i])
+}
+
+#' @rdname term_list
+#' @export
+print.term_list <- function(x, ...){
+  signstr <- ifelse(attr(x, "sign")>=0, "+", "-")
+  envstr <- sapply(attr(x, "env"), format)
+  termstr <- lapply(lapply(x, format), paste0, collapse="\n")
+
+  cat("Term List:\n")
+  cat(paste(signstr, termstr, envstr, collapse="\n"))
+  cat("\n")
+}
 
 .recurse_summation <- function(x, sign){
-  if(length(x)==1) {out <- list(x); attr(out,"sign")<-sign; out}
-  else if(length(x)==2 && x[[1L]]=="+") .recurse_summation(x[[2L]],sign)
-  else if(length(x)==2 && x[[1L]]=="-") .recurse_summation(x[[2L]],-sign)
-  else if(length(x)==3 && x[[1L]]=="+") {
-    l1 <- .recurse_summation(x[[2L]],sign)
-    l2 <- .recurse_summation(x[[3L]],sign)
-    out <- c(l1, l2)
-    attr(out,"sign") <- c(attr(l1,"sign"), attr(l2,"sign"))
-    out
-  }
-  else if(length(x)==3 && x[[1L]]=="-"){
-    l1 <- .recurse_summation(x[[2L]],sign)
-    l2 <- .recurse_summation(x[[3L]],-sign)
-    out <- c(l1, l2)
-    attr(out,"sign") <- c(attr(l1,"sign"), attr(l2,"sign"))
-    out
-  }
-  else if(x[[1L]]=="(") .recurse_summation(x[[2L]], sign)
-  else {out <- list(x); attr(out,"sign")<-sign; out}
+  if(length(x)==1)
+    term_list(x, sign)
+  else if(length(x)==2 && x[[1L]]=="+")
+    .recurse_summation(x[[2L]],sign)
+  else if(length(x)==2 && x[[1L]]=="-")
+    .recurse_summation(x[[2L]],-sign)
+  else if(length(x)==3 && x[[1L]]=="+")
+    c(.recurse_summation(x[[2L]],sign),
+      .recurse_summation(x[[3L]],sign))
+  else if(length(x)==3 && x[[1L]]=="-")
+    c(.recurse_summation(x[[2L]],sign),
+      .recurse_summation(x[[3L]],-sign))
+  else if(length(x[[1]]) && x[[1L]]=="(")
+    .recurse_summation(x[[2L]], sign)
+  else
+    term_list(x, sign)
 }
 
 
@@ -277,15 +387,16 @@ term.list.formula<-function(rhs, sign=+1){
 #' @describeIn formula.utilities
 #'
 #' \code{list_summands.call}, given an unevaluated call or expression
-#' containing the sum of one or more terms, returns a list of the
+#' containing the sum of one or more terms, returns an object of class [`term_list`] with the
 #' terms being summed, handling \code{+} and \code{-} operators and
 #' parentheses, and keeping track of whether a term has a plus or a
 #' minus sign.
 #'
-#' @return \code{list_summands.call} returns a list of unevaluated calls, with an additional numerical vector attribute \code{"sign"} with of the same length, giving the corresponding term's sign as \code{+1} or \code{-1}.
+#' @return \code{list_summands.call} returns an object of type
+#'   [`term_list`]; its `"env"` attribute is set to a list of
+#'   `NULL`s, however.
 #' 
 #' @export
-
 list_summands.call<-function(object){
   .recurse_summation(object, sign=+1)
 }
@@ -293,18 +404,39 @@ list_summands.call<-function(object){
 
 #' @describeIn formula.utilities
 #'
-#' \code{list_rhs.formula} returns a list containing terms in a given
-#' formula, handling \code{+} and \code{-} operators and parentheses, and
-#' keeping track of whether a term has a plus or a minus sign.
+#' \code{list_rhs.formula} returns an object of type [`term_list`],
+#' containing terms in a given formula, handling \code{+} and \code{-}
+#' operators and parentheses, and keeping track of whether a term has
+#' a plus or a minus sign.
 #'
 #' @return
-#' \code{list_rhs.formula} returns a list of formula terms, with an additional numerical vector attribute \code{"sign"} with of the same length, giving the corresponding term's sign as \code{+1} or \code{-1}.
+#' \code{list_rhs.formula} returns an object of type [`term_list`].
+#'
+#' @examples
+#' stopifnot(identical(list_rhs.formula(a~b),
+#'           structure(alist(b), sign=1, env=list(globalenv()), class="term_list")))
+#' stopifnot(identical(list_rhs.formula(~b),
+#'           structure(alist(b), sign=1, env=list(globalenv()), class="term_list")))
+#' stopifnot(identical(list_rhs.formula(~b+NULL),
+#'           structure(alist(b, NULL),
+#'                     sign=c(1,1), env=rep(list(globalenv()), 2), class="term_list")))
+#' stopifnot(identical(list_rhs.formula(~-b+NULL),
+#'           structure(alist(b, NULL),
+#'                     sign=c(-1,1), env=rep(list(globalenv()), 2), class="term_list")))
+#' stopifnot(identical(list_rhs.formula(~+b-NULL),
+#'           structure(alist(b, NULL),
+#'                     sign=c(1,-1), env=rep(list(globalenv()), 2), class="term_list")))
+#' stopifnot(identical(list_rhs.formula(~+b-(NULL+c)),
+#'           structure(alist(b, NULL, c),
+#'                     sign=c(1,-1,-1), env=rep(list(globalenv()), 3), class="term_list")))
+#'
 #' @export
 list_rhs.formula<-function(object){
   if (!is(object, "formula"))
     stop("Invalid formula of class ",sQuote(class(object)),".")
   
-  .recurse_summation(ult(object), sign=+1)
+  o <- .recurse_summation(ult(object), sign=+1)
+  structure(o, env = rep(list(environment(object)), length(o)))
 }
 
 
