@@ -1,12 +1,12 @@
-#  File R/misc.utilities.R in package statnet.common, part of the Statnet suite
-#  of packages for network analysis, https://statnet.org .
+#  File R/misc.utilities.R in package statnet.common, part of the
+#  Statnet suite of packages for network analysis, https://statnet.org .
 #
 #  This software is distributed under the GPL-3 license.  It is free,
 #  open source, and has the attribution requirements (GPL Section 7) at
-#  https://statnet.org/attribution
+#  https://statnet.org/attribution .
 #
-#  Copyright 2007-2019 Statnet Commons
-#######################################################################
+#  Copyright 2007-2022 Statnet Commons
+################################################################################
 #' reorder vector v into order determined by matching the names of its elements
 #' to a vector of names
 #' 
@@ -47,19 +47,19 @@ vector.namesmatch<-function(v,names,errname=NULL){
   v
 }
 
-
 #' "Compress" a data frame.
 #' 
-#' \code{compress.data.frame} "compresses" a data frame, returning unique rows
+#' \code{compress_rows.data.frame} "compresses" a data frame, returning unique rows
 #' and a tally of the number of times each row is repeated, as well as a
 #' permutation vector that can reconstruct the original data frame.
-#' \code{decompress.data.frame} reconstructs the original data frame.
+#' \code{decompress_rows.compressed_rows_df} reconstructs the original data frame.
 #' 
 #' 
-#' @param x For \code{compress.data.frame} a \code{\link{data.frame}} to be
-#' compressed. For \code{decompress.data.frame} a \code{\link{list}} as
-#' returned by \code{compress.data.frame}.
-#' @return For \code{compress.data.frame}, a \code{\link{list}} with three
+#' @param x For \code{compress_rows.data.frame} a \code{\link{data.frame}} to be
+#' compressed. For \code{decompress_rows.compress_rows_df} a \code{\link{list}} as
+#' returned by \code{compress_rows.data.frame}.
+#' @param ... Additional arguments, currently unused.
+#' @return For \code{compress_rows.data.frame}, a \code{\link{list}} with three
 #' elements: \item{rows }{Unique rows of \code{x}} \item{frequencies }{A vector
 #' of the same length as the number or rows, giving the number of times the
 #' corresponding row is repeated } \item{ordering}{A vector such that if
@@ -67,7 +67,7 @@ vector.namesmatch<-function(v,names,errname=NULL){
 #' equals the original data frame, except for row names} \item{rownames}{Row
 #' names of \code{x}}
 #' 
-#' For \code{decompress.data.frame}, the original data frame.
+#' For \code{decompress_rows.compressed_rows_df}, the original data frame.
 #' @seealso \code{\link{data.frame}}
 #' @keywords manip
 #' @examples
@@ -76,27 +76,28 @@ vector.namesmatch<-function(v,names,errname=NULL){
 #'                  V2=sample.int(2,30,replace=TRUE),
 #'                  V3=sample.int(4,30,replace=TRUE)))
 #' 
-#' (c <- compress.data.frame(x))
+#' (c <- compress_rows(x))
 #' 
-#' stopifnot(all(decompress.data.frame(c)==x))
+#' stopifnot(all(decompress_rows(c)==x))
+#'
 #' @export
-compress.data.frame<-function(x){
+compress_rows.data.frame<-function(x, ...){
   r <- rownames(x)
   o <- order.data.frame(x)
   x <- x[o, , drop=FALSE]
   firsts<-which(!duplicated(x))
   freqs<-diff(c(firsts,nrow(x)+1))
   x<-x[firsts, , drop=FALSE]
-  list(rows=x, frequencies=freqs, ordering=order(o), rownames=r) # Note that x[order(x)][order(order(x))]==x.
+  structure(x, frequencies=freqs, ordering=order(o), rownames=r, class=c("compressed_rows_df", class(x))) # Note that x[order(x)][order(order(x))]==x.
 }
 
-#' @rdname compress.data.frame
+#' @rdname compress_rows.data.frame
 #' @export
-decompress.data.frame<-function(x){
-  r <- x$rows
-  rn <- x$rownames
-  f <- x$frequencies
-  o <- x$ordering
+decompress_rows.compressed_rows_df<-function(x, ...){
+  r <- x
+  rn <- attr(x, "rownames")
+  f <- attr(x, "frequencies")
+  o <- attr(x, "ordering")
 
   out <- r[rep.int(seq_along(f), f),, drop=FALSE][o,, drop=FALSE]
   rownames(out) <- rn
@@ -723,6 +724,7 @@ persistEval <- function(expr, retries=NVL(getOption("eval.retries"), 5), beforeR
                                      is.pairlist(envir)) parent.frame() else baseenv(), verbose=FALSE){
   for(attempt in seq_len(retries)){
     out <- try(eval(expr, envir=envir, enclos=enclos), silent=TRUE)
+    #' @importFrom methods is
     if(!is(out, "try-error")) return(out)
     else{
       if(!missing(beforeRetry)) eval(beforeRetry, envir=envir, enclos=enclos)
@@ -745,3 +747,143 @@ persistEvalQ <- function(expr, retries=NVL(getOption("eval.retries"), 5), before
 
   persistEval(expr=expr, retries=retries, beforeRetry=beforeRetry, envir=envir, enclos=enclos, verbose=verbose)
 }
+
+#' Truncate values of high magnitude in a vector.
+#'
+#' @param x a numeric or integer vector.
+#' @param replace a number or a string `"maxint"` or `"intmax"`.
+#'
+#' @return Returns `x` with elements whose magnitudes exceed `replace`
+#'   replaced replaced by `replace` (or its negation). If `replace` is
+#'   `"maxint"` or `"intmax"`, `.Machine$integer.max` is used instead.
+#'
+#' `NA` and `NAN` values are preserved.
+#'
+#' @export
+deInf <- function(x, replace=1/.Machine$double.eps){
+  NVL(x) <- integer(0)
+  if(tolower(replace) %in% c("maxint","intmax")) replace <- .Machine$integer.max
+  ifelse(is.nan(x) | abs(x)<replace, x, sign(x)*replace)
+}
+
+#' A [split()] method for [`array`] and [`matrix`] types on a margin.
+#'
+#' These methods split an [`array`] and [`matrix`] into a list of
+#' arrays or matrices with the same number of dimensions
+#' according to the specified margin.
+#'
+#' @param x A [`matrix`] or an [`array`].
+#' @param f,drop See help for [split()]. Note that `drop` here is
+#'   *not* for array dimensions: these are always preserved.
+#' @param margin Which margin of the array to split along. `NULL`
+#'   splits as [`split.default`], dropping dimensions.
+#' @param ... Additional arguments to [split()].
+#'
+#' @examples
+#'
+#' x <- diag(5)
+#' f <- rep(1:2, c(2,3))
+#' split(x, f, margin=1) # Split rows.
+#' split(x, f, margin=2) # Split columns.
+#'
+#' # This is similar to how data frames are split:
+#' stopifnot(identical(split(x, f, margin=1),
+#'           lapply(lapply(split(as.data.frame(x), f), as.matrix), unname)))
+#'
+#' @export
+split.array <- function(x, f, drop = FALSE, margin = NULL, ...){
+  if(is.null(margin)) return(NextMethod("split"))
+  d <- dim(x)
+  margin <- as.integer(margin)
+  if(margin < 1L || margin > length(d)) stop(sQuote("margin"), " must be between 1 and the dimensionality of ", sQuote("x"), ".")
+
+  args <- c(list(x), rep(TRUE, length(d)), list(drop=FALSE))
+  ind_call <- function(ind){
+    args[[margin+1L]] <- ind
+    do.call(`[`, args)
+  }
+  lapply(split(x = seq_len(dim(x)[margin]), f = f, drop = drop, ...), ind_call)
+}
+
+#' @rdname split.array
+#' @export
+split.matrix <- split.array
+
+#' Convert a list to an atomic vector if it consists solely of atomic elements of length 1.
+#'
+#' This behaviour is not dissimilar to that of [simplify2array()], but
+#' it offers more robust handling of empty or NULL elements and never
+#' promotes to a matrix or an array, making it suitable to be a column
+#' of a [`data.frame`].
+#'
+#' @param x an R [`list`] to be simplified.
+#' @param toNA a character string indicating whether `NULL` entries
+#'   (if `"null"`) or 0-length entries including `NULL` (if `"empty"`)
+#'   should be replaced with `NA`s before attempting conversion;
+#'   specifying `keep` or `FALSE` leaves them alone (typically
+#'   preventing conversion).
+#' @param empty a character string indicating how empty lists should
+#'   be handled: either `"keep"`, in which case they are unchanged or
+#'   `"unlist"`, in which cases they are unlisted (typically to
+#'   `NULL`).
+#' @param ... additional arguments passed to [unlist()].
+#'
+#' @return an atomic vector or a list of the same length as `x`.
+#' @examples
+#'
+#' (x <- as.list(1:5))
+#' stopifnot(identical(simplify_simple(x), 1:5))
+#'
+#' x[3] <- list(NULL) # Put a NULL in place of 3.
+#' x
+#' stopifnot(identical(simplify_simple(x, FALSE), x)) # Can't be simplified without replacing the NULL.
+#'
+#' stopifnot(identical(simplify_simple(x), c(1L,2L,NA,4L,5L))) # NULL replaced by NA and simplified.
+#'
+#' x[[3]] <- integer(0)
+#' x
+#' stopifnot(identical(simplify_simple(x), x)) # A 0-length vector is not replaced by default,
+#' stopifnot(identical(simplify_simple(x, "empty"), c(1L,2L,NA,4L,5L))) # but can be.
+#'
+#' (x <- lapply(1:5, function(i) c(i,i+1L))) # Elements are vectors of equal length.
+#' simplify2array(x) # simplify2array() creates a matrix,
+#' stopifnot(identical(simplify_simple(x), x)) # but simplify_simple() returns a list.
+#'
+#' @export
+simplify_simple <- function(x, toNA = c("null","empty","keep"), empty = c("keep", "unlist"), ...){
+  if(isFALSE(toNA)) toNA <- "keep"
+  toNA <- match.arg(toNA)
+  empty <- match.arg(empty)
+
+  if(is.atomic(x)) return(x)
+
+  x <- switch(toNA,
+              keep = x,
+              null = lapply(x, NVL, NA),
+              empty = lapply(x, EVL, NA))
+
+  if(length(x)==0) switch(empty, keep=x, unlist=unlist(x, recursive=FALSE, ...))
+  else if(all(lengths(x)==1L) && all(vapply(x, is.atomic, logical(1)))) unlist(x, recursive=FALSE, ...)
+  else x
+}
+
+#' A wrapper for base::attr which defaults to exact matching.
+#'
+#' @param x,which,exact as in \code{base::attr}, but with \code{exact}
+#'   defaulting to \code{TRUE} in this implementation
+#'
+#' @return as in \code{base::attr}
+#' @examples
+#'
+#' x <- list()
+#' attr(x, "name") <- 10
+#'
+#' base::attr(x, "n")
+#'
+#' stopifnot(is.null(attr(x, "n")))
+#'
+#' base::attr(x, "n", exact = TRUE)
+#' @export
+attr <- function(x, which, exact = TRUE) {
+  base::attr(x, which, exact)
+}
